"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateSchemaFromSimpleSchema = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const index_js_1 = require("../core/index.js");
const schemaTypes_js_1 = require("../schemaTypes.js");
const schemaFactory_js_1 = require("./schemaFactory.js");
const schemaFactoryAlpha_js_1 = require("./schemaFactoryAlpha.js");
const factory = new schemaFactoryAlpha_js_1.SchemaFactoryAlpha(undefined);
/**
 * Create {@link FieldSchema} from a SimpleTreeSchema.
 * @remarks
 * Only use this API if hand written schema (produced using {@link SchemaFactory} cannot be provided.
 *
 * Using generated schema with schema aware APIs (designed to work with strongly typed schema) like {@link TreeViewConfiguration}
 * will produce a poor TypeScript typing experience which is subject to change.
 *
 * Editing through a view produced using this schema can easily violate invariants other users of the document might expect and must be done with great care.
 * @alpha
 */
function generateSchemaFromSimpleSchema(simple) {
    const context = new Map([...simple.definitions].map(([id, schema]) => [
        id,
        // This relies on the caching in evaluateLazySchema so that it only runs once.
        () => generateNode(id, schema, context),
    ]));
    return generateFieldSchema(simple, context);
}
exports.generateSchemaFromSimpleSchema = generateSchemaFromSimpleSchema;
function generateFieldSchema(simple, context) {
    const allowed = generateAllowedTypes(simple.allowedTypesIdentifiers, context);
    // Using createFieldSchema could work, but would require setting up the default providers.
    switch (simple.kind) {
        case schemaTypes_js_1.FieldKind.Identifier:
            return schemaFactoryAlpha_js_1.SchemaFactoryAlpha.identifier({ metadata: simple.metadata });
        case schemaTypes_js_1.FieldKind.Optional:
            return schemaFactoryAlpha_js_1.SchemaFactoryAlpha.optional(allowed, { metadata: simple.metadata });
        case schemaTypes_js_1.FieldKind.Required:
            return schemaFactoryAlpha_js_1.SchemaFactoryAlpha.required(allowed, { metadata: simple.metadata });
        default:
            return (0, internal_1.unreachableCase)(simple.kind);
    }
}
function generateAllowedTypes(allowed, context) {
    return [...allowed].map((id) => context.get(id) ?? (0, internal_1.fail)(0xb5a /* Missing schema */));
}
function generateNode(id, schema, context) {
    switch (schema.kind) {
        case index_js_1.NodeKind.Object: {
            const fields = {};
            for (const [key, field] of schema.fields) {
                fields[key] = generateFieldSchema(field, context);
            }
            return factory.object(id, fields, { metadata: schema.metadata });
        }
        case index_js_1.NodeKind.Array:
            return factory.arrayAlpha(id, generateAllowedTypes(schema.allowedTypesIdentifiers, context), { metadata: schema.metadata });
        case index_js_1.NodeKind.Map:
            return factory.mapAlpha(id, generateAllowedTypes(schema.allowedTypesIdentifiers, context), { metadata: schema.metadata });
        case index_js_1.NodeKind.Leaf:
            return (schemaFactory_js_1.SchemaFactory.leaves.find((leaf) => leaf.identifier === id) ??
                (0, internal_1.fail)(0xb5b /* Missing schema */));
        default:
            return (0, internal_1.unreachableCase)(schema);
    }
}
//# sourceMappingURL=schemaFromSimple.js.map